from neuron import h
from matplotlib import pyplot, gridspec

h.load_file('stdrun.hoc')


class Cell:
    '''Integrate and Fire cell'''
    def __init__(self, _id, interval, tau=10):
        self.id = _id
        self.dynamics = h.IntIbFire()
        self._netcons = []
        self._spike_detector = h.NetCon(self.dynamics, None)
        self._spike_times = h.Vector()
        self._spike_detector.record(self._spike_times)
        self.dynamics.tau = tau        # ms
        self.dynamics.invl = interval  # ms

    def connect(self, other, weight=-0.3, delay=4):
        '''connects self -> other'''
        if isinstance(other, Cell):
            connection = h.NetCon(self.dynamics, other.dynamics)
        else:
            connection = h.NetCon(self.dynamics, other)
        connection.weight[0] = weight
        connection.delay = delay
        self._netcons.append(connection)
        return connection


class Network:
    '''a network'''
    def __init__(
            self,
            num_cells=100,
            interval_low=10,
            interval_high=15,
            weight=-0.005,
            delay=10,
            tau=10):
        '''create a network of size num_cells'''
        dinterval = (interval_high - interval_low) / float(num_cells)
        self.cells = [
            Cell(i, interval_low + i * dinterval, tau) for i in range(num_cells)
        ]
        self.wire(weight, delay)

    def wire(self, weight=-0.005, delay=10):
        '''all-to-all (non-self) connectivity'''
        for pre in self.cells:
            for post in self.cells:
                if pre != post:
                    pre.connect(post, weight, delay)
    
    def set_weight(self, new_weight):
        '''set all the synaptic weights'''
        for cell in self.cells:
            for nc in cell._netcons:
                nc.weight[0] = new_weight

    def set_delay(self, new_delay):
        '''set all the synaptic delays'''
        for cell in self.cells:
            for nc in cell._netcons:
                nc.delay = new_delay
    
    weight = property(fset=set_weight, doc='synaptic weights')
    delay = property(fset=set_delay, doc='synaptic delay')


def plot_raster(ax, network):
    '''uses pyplot to plot the spike times.'''
    for cell in network.cells:
        if cell._spike_times:
            ax.vlines(cell._spike_times, cell.id - 0.45, cell.id + 0.45)

h.cvode_active(True)


def create_readout_cell(network, weight=1e-5, delay=4, tau=5):
    output_cell = h.Section(name='output_cell')
    output_cell.L = output_cell.diam = 10
    output_cell.insert('hh')
    syns = []
    ncs = []
    for cell in network.cells:
        syn = h.ExpSyn(output_cell(0.5))
        syn.e = 0
        syn.tau = tau
        nc = h.NetCon(cell.dynamics, syn)
        nc.weight[0] = weight
        nc.delay = delay
        syns.append(syn)
        ncs.append(nc)
    return output_cell, (syns, ncs)


def plot_raster_and_output_mv(network, t, output_v):
    '''plot a raster plot and the readout cell timecourse'''
    fig = pyplot.figure()
    gs = gridspec.GridSpec(3, 1)
    ax = fig.add_subplot(gs[0:2, :])
    plot_raster(ax, network)
    ax.set_ylabel('cell id')
    ax.set_xticks([])
    ax.set_xlim([0, h.t])
    ax.set_ylim([-0.5, len(network.cells) - 0.5])
    ax = fig.add_subplot(gs[2, :])
    ax.plot(t, output_v)
    ax.set_xlim([0, h.t])
    ax.set_xlabel('t (ms)')
    ax.set_ylabel('output cell (mV)')
    pyplot.show()


if __name__ == '__main__':
    network = Network(
        num_cells=100,
        interval_low=10,
        interval_high=15,
        delay=4,
        weight=-0,
        tau=10)
    
    output_sec, output_conns = create_readout_cell(network, weight=5e-6)
    
    output_v = h.Vector()
    output_v.record(output_sec(0.5)._ref_v)
    t = h.Vector()
    t.record(h._ref_t)

    h.finitialize(-65)
    h.continuerun(1000)
    plot_raster_and_output_mv(network, t, output_v)

    '''
    can run other experiments after this is done using the network
    
    e.g.

    network.weight = -0.005
    h.finitialize(-65)
    h.continuerun(1000)
    plot_raster_and_output_mv(network, t, output_v)
    '''

